
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <pthread.h>

#include "xstuff.c"
#include "mazeroutines__.c"


// ============= Declare constants ==============
#define WALL_LIMIT 44000
#define PI 3.14159265358

#define KEY_Z 0
#define KEY_X 1
#define KEY_Q 2
#define KEY_V 3
#define KEY_SPACE 4


// ============= Declare global variables ==============

int key[255];

struct MazeGameData {

 int orix,oriy,walls;

 float distmult;
 float dists_a[WALL_LIMIT],angles_a[WALL_LIMIT],dists_b[WALL_LIMIT],angles_b[WALL_LIMIT];
 float ballx,bally,oballx,obally,offset,olset;

 float xmomentum,ymomentum;
 float startangle,startdist,goalangle,goaldist;
 int pxx,pyy;
 int ballsize;
 int holesize;
 int ball_screen_x,ball_screen_y;

 int view_style;

 float sizeadjust;

};

float FNgetangl(struct MazeGameData *gd,float x,float y);
float FNgetdistanc(struct MazeGameData *gd,float x,float y);
int FNget(int x, int y, struct Maze *m);
void PrepareMazeGame(struct MazeGameData *gd, struct Maze *m );
int PlayGame(struct MazeGameData *gd, struct Maze *m);
int PROCquit(void);
int iABS(int in);
float fSGN(float in);
void PROCcollision(struct MazeGameData *gd, struct Maze *m);
void PROCspinanimation(struct MazeGameData *gd,struct Maze *m);
void PROCgoalanimation(struct MazeGameData *gd,struct Maze *m);
void PROCredrawmaze(struct MazeGameData *gd);
void PROCredrawball(struct MazeGameData *gd);

// =====================================================


int main(int argc, char **argv){

 XRANDrand=(unsigned long int) time(NULL);
 XRANDranb=(unsigned long int) (~time(NULL))>>1;

 int i; for(i=0; i<256; i++){ key[i]=0; }

#if 0
 MyInit(640,480);
 MyHandleEvents();
 Circle(20,20,20);
 Rectangle(50,50,20,20);
 Gcol(255,128,128);
 RectangleFill(70,70,30,30);
 MyHandleEvents_WithBlocking();
 return 0;
#endif

 int s;
 if(argc>1)
  s=atoi(argv[1]);
 else
  s=0;

 MyInit(640,480);
 XStoreName(Mydisplay,Mywindow, "Magic Maze - Press Q to quit");
 //XSynchronize(Mydisplay, True);

 struct Maze *maze = CreateMaze(512);
 struct MazeGameData *game = calloc(1,sizeof(struct MazeGameData));

 if(s){
  PrepareMaze(maze,s);
  PrepareMazeGame( game, maze );
  printf("Starting the game\n");
  PlayGame( game, maze);
  MyCleanup();
  printf("Exiting\n");
  exit(0);
 }
 
 int size=0;
 for(size=14; size<=400; size+=4){
  PrepareMaze(maze,size);
  PrepareMazeGame( game, maze );
  printf("Starting the game\n");
  PlayGame( game, maze);
 }

 return 0;
}


float FNgetangl(struct MazeGameData *gd,float x,float y){
 return atan2(gd->oriy-y,gd->orix-x);
}

float FNgetdistanc(struct MazeGameData *gd,float x,float y){
 float dx,dy;
 dx=(gd->orix-x)*(gd->orix-x);
 dy=(gd->oriy-y)*(gd->oriy-y);
 return sqrt(dx+dy);
}

int FNget(int x, int y, struct Maze *m){
 if( x<0 || y<0 || x>m->MAZEMAX || y>m->MAZEMAX) return -1;
 return m->maze[x][y];
}

void PrepareMazeGame(struct MazeGameData *gd, struct Maze *m ){
 char **maze = m->maze;
 m->MAZEMAX = m->MAZEMAX & -2;
 int MAZEMAX = m->MAZEMAX; 
 int i,j,state=0,blk=0,count=0,wallp=0,xa,ya;
  
 gd->distmult=(float)(480.0/3)/(float)(MAZEMAX/2);
 gd->orix=MAZEMAX/2;
 gd->oriy=gd->orix;

 state=0;count=0;
 //check for horizontal walls
 for(j=0;j<=MAZEMAX+1;j+=1){
  for(i=0;i<=MAZEMAX+1;i+=1){
   blk=FNget(i,j,m);

   if(state==0){

    if(blk==mazeroutines_SVAL){
     state=1;
     count=0;
     xa=i;
     ya=j;
    }

   }else{

    if(blk != mazeroutines_SVAL){ 
     state=0;
     if(count>1){
      gd->angles_a[wallp]= FNgetangl(gd,xa,ya);
      gd->dists_a[wallp] = FNgetdistanc(gd,xa,ya)*gd->distmult;
      gd->angles_b[wallp]= FNgetangl(gd,i-1,j);
      gd->dists_b[wallp] = FNgetdistanc(gd,i-1,j)*gd->distmult;
      wallp+=1;
     }
    }

   }

   count+=1;
  }
 }

 state=0;count=0;
 //check for vertical walls
 for(j=0;j<=MAZEMAX+1;j+=1){
  for(i=0;i<=MAZEMAX+1;i+=1){
   blk=FNget(j,i,m);

   if(state==0){

    if(blk==mazeroutines_SVAL){
     state=1;
     count=0;
     xa=j;
     ya=i;
    }

   }else{

    if(blk != mazeroutines_SVAL){ 
     state=0;
     if(count>1){
      gd->angles_a[wallp]= FNgetangl(gd,xa,ya);
      gd->dists_a[wallp] = FNgetdistanc(gd,xa,ya)*gd->distmult;
      gd->angles_b[wallp]= FNgetangl(gd,j,i-1);
      gd->dists_b[wallp] = FNgetdistanc(gd,j,i-1)*gd->distmult;
      wallp+=1;
     }
    }

   }

   count+=1;
  }
 }
 
 gd->walls = wallp-1;
}

#define SpeedAdjust 0.7

#define movepower ((0.15/5)*SpeedAdjust)
#define momentumcap (0.45*SpeedAdjust)
#define controlamount (((PI*2)/64)*SpeedAdjust)

#define centrefix 0.5

void SetScreenDimensions(struct MazeGameData *gd, struct Maze *m){
 gd->startangle=FNgetangl(gd,1,0); gd->startdist=FNgetdistanc(gd,1,0); gd->goalangle=FNgetangl(gd,m->MAZEMAX-1,m->MAZEMAX);
 gd->goaldist=FNgetdistanc(gd,m->MAZEMAX-1,m->MAZEMAX);
 gd->pxx=WinW/2;
 gd->pyy=WinH/2;
 int bsconst=(int)((float)480.0*2)/3.0;
 int hsconst=(int)((float)480.0*2)/(PI/2);
 gd->ballsize=(int)(float)bsconst/(float)m->MAZEMAX;
 gd->holesize=(int)(float)hsconst/(float)m->MAZEMAX;
 gd->sizeadjust = WinH/462.5;
 //system("Shia");
}

// ============================================================================================================================================================
// ============================================================================================================================================================
// ============================================================================================================================================================
// ============================================================================================================================================================

int Maze_HandleEvents(struct MazeGameData *gd, struct Maze *m){
 int i,screenchange=0;

 while(1){

  if( XCheckMaskEvent( Mydisplay,ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask | StructureNotifyMask, &Myevent )==False ) break;

  //printf("event.type: %d\n",Myevent.type);

  switch(Myevent.type){
  case Expose:
   //Cls();
   break;
  case ConfigureNotify:
   screenchange=1;
   WinW = Myevent.xconfigure.width;
   WinH = Myevent.xconfigure.height;
   usleep(5111);
   break;
  case ButtonPress:
   while( Myevent.type != ButtonRelease ) XNextEvent(Mydisplay, &Myevent);
   break;
  case ButtonRelease:
   //system("Shia");
   MyCleanup();
   exit(0);
   break;
  case KeyPress:
   key[Myevent.xkey.keycode]=1;
   //printf("press   %d\n",Myevent.xkey.keycode);
   if( Myevent.xkey.keycode == 24 ){
    printf("Quit: Q pressed\n");
    MyCleanup();
    exit(0);
   }
   break;
  case KeyRelease:
   //printf("release %d\n",Myevent.xkey.keycode);
   key[Myevent.xkey.keycode]=0;
   break;
  }//endcase

 }//endwhile

 if( screenchange ){ 
  SetScreenDimensions(gd,m);
  XFreePixmap(Mydisplay,Mypixmap);
  Mypixmap = XCreatePixmap(Mydisplay,Mywindow,WinW,WinH,SDepth);
  Mydrawable=Mypixmap;
  //printf("screenchange \n");
  //XFlush(Mydisplay);
  Cls();
  //printf("screenchange x \n");
 }
 
}//endproc

// ============================================================================================================================================================
// ============================================================================================================================================================
// ============================================================================================================================================================
// ============================================================================================================================================================

#define UseMazeCls 1
Maze_Cls(){
 XSetForeground(Mydisplay, MyGC,BlackPixel(Mydisplay,Myscreen) );
 XFillRectangle(Mydisplay,Mydrawable,MyGC,(WinW*0.5)-WinH*1.15*0.5,0, WinH*1.15,WinH);
 XSetForeground(Mydisplay, MyGC,MyOldCol );
}

int PlayGame(struct MazeGameData *gd, struct Maze *m){
 gd->ballx=1; gd->bally=0; gd->oballx=2; gd->obally=2; gd->offset=PI; gd->olset=1;
 gd->xmomentum=0; gd->ymomentum=0;

 SetScreenDimensions(gd,m); Cls();

 while(!(  ( ((int)gd->oballx)==(m->MAZEMAX-1) || ((int)gd->oballx)==(m->MAZEMAX-2) ) && ((int)gd->obally)==(m->MAZEMAX)) ){

  //printf("x,y:\t%f\t%f\n",gd->ballx,gd->bally);
#if UseMazeCls
  Maze_Cls();
#else
  Cls();
#endif
  //if(gd->offset!=gd->olset) 
  PROCredrawmaze(gd);
  PROCredrawball(gd);

  PROCcollision(gd,m);

  gd->olset=gd->offset;
  if(key[65]){
   PROCspinanimation(gd,m);
  }else if(key[52]){
   gd->offset-=controlamount;
  }else if(key[53]){
   gd->offset+=controlamount;
  }/*else if(key[KEY_V]){
   gd->view_style=!gd->view_style;
  }*/
/*
  offset+=(((float) myaxes[0])/261112);
  if(mybuttons[0]) PROCspinanimation();
  if(mybuttons[1]) {
   view_style=!view_style;
   mybuttons[1]=0;
  }
  if(mybuttons[2]) {
   rest(40);
  }
*/
  // ========== TEMP TEST ==============
  //gd->offset += movepower/3;
  // ===================================

  //PROCwait(23);
  //XFlush(Mydisplay);
  MyCopy();
  Maze_HandleEvents(gd,m);
  usleep(16666);
 }
 PROCgoalanimation(gd,m);
}

int PROCquit(void){
 exit(0); 
}

int iABS(int in){
 if(in<0) return in*-1;
 return in;
}
float fSGN(float in){
 if(in>0) return 1;
 if(in<0) return -1;
 return 0;
}

void PROCcollision(struct MazeGameData *gd, struct Maze *m){

 gd->oballx=gd->ballx;
 gd->obally=gd->bally;
 gd->xmomentum-=cos(gd->offset-(PI/2))*movepower;
 gd->ymomentum+=sin(gd->offset-(PI/2))*movepower;
 
 if(fabs(gd->xmomentum)>momentumcap) gd->xmomentum=momentumcap*fSGN(gd->xmomentum);
 if(fabs(gd->ymomentum)>momentumcap) gd->ymomentum=momentumcap*fSGN(gd->ymomentum);

 gd->ballx += gd->xmomentum;
 gd->bally += gd->ymomentum;


 float sx,sy, ex,ey, mx,my;
 sx = gd->oballx; sy = gd->obally;
 ex = gd->ballx;  ey = gd->bally;
 mx = (sx+ex)/2;  my=(sy+ey)/2;
 int check_a,check_b,ccheck_a,ccheck_b;
 
 check_a = iABS(FNget(gd->oballx + centrefix,gd->bally + centrefix,m));
 check_b = iABS(FNget(gd->ballx + centrefix,gd->obally + centrefix,m));

 if( check_a ){
  do{
   my=(sy+ey)/2;
   ccheck_a=iABS(FNget(gd->oballx + centrefix, my + centrefix,m));
   if(ccheck_a){
    ey=my;
   } else {
    sy=my;
   }
  }while( fabs(sy-ey)>0.01 );
  gd->bally=sy;
 }

 if( check_b ){
  do{
   mx=(sx+ex)/2;
   ccheck_b=iABS(FNget(mx + centrefix,gd->bally + centrefix,m));
   if(ccheck_b){
    ex=mx;
   } else {
    sx=mx;
   }
  }while( fabs(sx-ex)>0.01 );
  gd->ballx=sx;
 }

 // stuckfix
 if ( iABS(FNget(gd->ballx + centrefix,gd->bally + centrefix,m)) && !(gd->ballx == gd->oballx && gd->bally == gd->obally) ){
  printf("stuckfix activated\n");
  //system("beep -f 200 -l 10 &");
  float xd,yd;
  xd=(gd->ballx - gd->oballx)/100;
  yd=(gd->bally - gd->obally)/100;
  while( iABS(FNget(gd->ballx + centrefix,gd->bally + centrefix,m)) ) {
   gd->ballx -= xd;
   gd->bally -= yd;
  }
 }

 // reduce momentum when the ball is touching a wall
 if(check_a ) gd->ymomentum = gd->ymomentum*0.7;
 if(check_b ) gd->xmomentum = gd->xmomentum*0.7;
 // drag
 if(!check_a) gd->ymomentum = gd->ymomentum*0.999;
 if(!check_b) gd->xmomentum = gd->xmomentum*0.999;

}

void PROCspinanimation(struct MazeGameData *gd,struct Maze *m){
 int i,steps; float movestep;
 steps=16;
 movestep=PI/steps;
 for(i=1;i<=steps;i+=1){
  gd->offset -= movestep;
#if UseMazeCls
  Maze_Cls();
#else
  Cls();
#endif
  PROCredrawmaze(gd);
  PROCredrawball(gd);
  XFlush(Mydisplay);
  MyCopy();
  XFlush(Mydisplay);
  usleep(16666);
 }
}


void PROCgoalanimation(struct MazeGameData *gd,struct Maze *m){
 int i;
 float scale=0.9575;
 do{
#if UseMazeCls
  Maze_Cls();
#else
  Cls();
#endif
  Gcol(255,255,255);
  PROCredrawmaze(gd);
  MyCopy();
  Maze_HandleEvents(gd,m);
  usleep(16666);
  
  gd->sizeadjust *= scale;
  gd->offset += movepower;
  
 }while(gd->sizeadjust > 0.0025);
}

void PROCredrawmaze(struct MazeGameData *gd){
 int i;
 for(i=0;i<=gd->walls;i+=1){
  Line(gd->pxx + cos(gd->angles_a[i]+gd->offset)*gd->dists_a[i] *gd->sizeadjust, gd->pyy + sin(gd->angles_a[i]+gd->offset)*gd->dists_a[i] *gd->sizeadjust,
               gd->pxx + cos(gd->angles_b[i]+gd->offset)*gd->dists_b[i]*gd->sizeadjust, gd->pyy + sin(gd->angles_b[i]+gd->offset)*gd->dists_b[i]*gd->sizeadjust);
 }
 Gcol(128,0,255);
 Circle(gd->pxx + (cos(gd->startangle+gd->offset) * gd->startdist * gd->distmult * gd->sizeadjust),
        gd->pyy + (sin(gd->startangle+gd->offset) * gd->startdist * gd->distmult * gd->sizeadjust),
        gd->holesize * gd->sizeadjust);
 Gcol(0,255,0);
 Circle(gd->pxx + (cos(gd->goalangle+gd->offset) * gd->goaldist * gd->distmult * gd->sizeadjust),
        gd->pyy + (sin(gd->goalangle+gd->offset) * gd->goaldist * gd->distmult * gd->sizeadjust),
        gd->holesize * gd->sizeadjust);
}

void PROCredrawball(struct MazeGameData *gd){

 int ball_screen_x = (int)gd->pxx + cos(FNgetangl(gd,gd->ballx,gd->bally)+gd->offset) * FNgetdistanc(gd,gd->ballx,gd->bally) * gd->distmult * gd->sizeadjust;
 int ball_screen_y = (int)gd->pyy + sin(FNgetangl(gd,gd->ballx,gd->bally)+gd->offset) * FNgetdistanc(gd,gd->ballx,gd->bally) * gd->distmult * gd->sizeadjust;
  
 Gcol(255,255,255);
 CircleFill(ball_screen_x,ball_screen_y,gd->ballsize * gd->sizeadjust);
}

